/* flow.c - Bulthoff, Little and Poggio optical flow */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "pnmio.h"
#include "getargs.h"


#define		ABS(x)		(((x) >= 0) ? (x) : -(x))
#define		SQR(x)		((x) * (x))
#define		MAX(x, y)	(((x) > (y)) ? (x) : (y))


#define		UCHAR		unsigned char

#define		YES		-1
#define		NO		0



/* function prototypes */
void flow( UCHAR *im1, UCHAR *im2, char *vx, char *vy, int nr, int nc, int wsize, float sigma, float bw );
void vflow( UCHAR *im1, UCHAR *im2, char *vx, char *vy, UCHAR *mask, int nr, int nc,
			int wsize, float sigma, float bw );
void verify( char *vx, char *vy, char *vx_b, char *vy_b, UCHAR *mask, int nr, int nc );
void match( UCHAR *im1, UCHAR *im2, UCHAR *match, int nr, int nc, int xoffset, int yoffest );
void smooth( UCHAR *in, UCHAR *out, int nr, int nc, int size );
void setmax( UCHAR *mtchim, UCHAR *maxim, char *vx, char*vy, int nr, int nc, int xoffset, int yoffset );
void setmax_b( UCHAR *mtchim, UCHAR *maxim, char *vx, char*vy, int nr, int nc, int xoffset, int yoffset );
void scale( UCHAR *im, int nr, int nc );
void magnitude( char *vx, char *vy, UCHAR *mag, int nr, int nc );
void gaussian( UCHAR *in, UCHAR *out, int nr, int nc, float sigma );
float normal( float x, float mu, float sigma );
void dog( UCHAR *in, char *out, int nr, int nc, float sigma );
void LogOfImage( UCHAR *in, UCHAR *out, int nr, int nc );

void DericheSmooth( UCHAR *in, UCHAR *out, int nr, int nc, float sigma );


int wsize = 5;
float sigma = 5.0;
float dogbw = 2.0;
int start = 0;
int stop = 5;
char *input = "input%03d.pgm";
char *xout = "xout%03d.pgm";
char *yout = "yout%03d.pgm";
char *maskout = "mask%03d.pgm";
int vfy = NO;
int dolog = NO;

ARG arglist[] = {
	{	"width",	INTEGER,	&wsize,		"motion window size"	},
	{	"smooth",	FLOAT,		(int *)&sigma,	"smoothing sigma"	},
	{	"dog",		FLOAT,		(int *)&dogbw,	"dog sigma"		},
	{	"start",	INTEGER,	&start,		"start image index"	},
	{	"stop",		INTEGER,	&stop,		"stop image index"	},
	{	"input",	STRING,		(int *)&input,	"input name format"	},
	{	"x",		STRING,		(int *)&xout,	"x motion name format"	},
	{	"y",		STRING,		(int *)&yout,	"y motion name format"	},
	{	"mask",		STRING,		(int *)&maskout,"mask name format"	},
	{	"v",		BOOLEAN,	(int *)&vfy,	"perform verified flow" },
};

#define		NARGS		(sizeof( arglist ) / sizeof( ARG ))


main( int argc, char **argv )
{
int nr, nc, n, nr2, nc2;
UCHAR *image1, *image2, *mask;
char *vx, *vy;
char inname1[100], inname2[100], outname[100];
int i;

	/* check command line args */
	argc = getargs( argc, argv, arglist, NARGS );
	if ( argc != 1 ) {
		fprintf( stderr, "Usage: %s [switches]", argv[0] );
		pr_usage( arglist, NARGS );
		exit( 0 );
	}

	if ( stop < start ) {
		fprintf( stderr, "Error:  stop index must be >= start index\n" );
		exit( 0 );
	}

	/* read first image to get image size */
	sprintf( inname1, input, start );
	if ( pgm_read( inname1, &image1, &nr, &nc ) ) {
		fprintf( stderr, "%s:  failed to read image <%s>\n", argv[0], inname1 );
		exit( 0 );
	}
	free( image1 );

	n = nr * nc;
	vx = (char *)malloc( n * sizeof( char ) );
	vy = (char *)malloc( n * sizeof( char ) );
	mask = (UCHAR *)malloc( n * sizeof( UCHAR ) );

	for( i = start; i < stop; i++ ) {
		/* read in first image */
		sprintf( inname1, input, i );
		if ( pgm_read( inname1, &image1, &nr2, &nc2 ) ) {
			fprintf( stderr, "%s:  failed to read image <%s>\n", argv[0], inname1 );
			exit( 0 );
		}
		if ( nr != nr2 || nc != nc2 ) {
			fprintf( stderr, "%s:  image dimensions do not match\n", argv[0] );
			exit( 0 );
		}

		/* read in second image and make sure that dimensions match */
		sprintf( inname2, input, i+1 );
		if ( pgm_read( inname2, &image2, &nr2, &nc2 ) ) {
			fprintf( stderr, "%s:  failed to read image <%s>\n", argv[0], inname2 );
			exit( 0 );
		}
		if ( nr != nr2 || nc != nc2 ) {
			fprintf( stderr, "%s:  image dimensions do not match\n", argv[0] );
			exit( 0 );
		}
		if ( vfy )
			fprintf( stderr, "Computing verified flow for %s to %s\n", inname1, inname2 );
		else
			fprintf( stderr, "Computing flow for %s to %s\n", inname1, inname2 );


		if ( dolog ) {
			LogOfImage( image1, image1, nr, nc );
			pgm_write( "booger.log.pgm", image1, nr, nc );
			LogOfImage( image2, image2, nr, nc );
		}

		if ( vfy )
			vflow( image1, image2, vx, vy, mask, nr, nc, wsize, sigma, dogbw );
		else
			flow( image1, image2, vx, vy, nr, nc, wsize, sigma, dogbw );

		sprintf( outname, xout, i );
		pgm_write( outname, vx, nr, nc );
		sprintf( outname, yout, i );
		pgm_write( outname, vy, nr, nc );
		if ( vfy ) {
			sprintf( outname, maskout, i );
			pgm_write( outname, mask, nr, nc );
		}

		free( image2 );
		free( image1 );
	}
}

/* unverified optical flow */

void flow( UCHAR *im1, UCHAR *im2, char *vx, char *vy, int nr, int nc, int wsize, float sigma, float bw )
{
UCHAR *mtchim, *maxim;
int x, y;

	dog( im1, im1, nr, nc, bw );
	dog( im2, im2, nr, nc, bw );

	mtchim = (UCHAR *)malloc( nr * nc * sizeof(UCHAR) );
	maxim  = (UCHAR *)malloc( nr * nc * sizeof(UCHAR) );

	/* initialize flow to zero offest in both directions */
	match( im1,im2, maxim, nr, nc, 0, 0 );		/* matching */
	/* smooth( maxim, maxim, nr, nc, smsize );		/* smooth matches */
	DericheSmooth( maxim, maxim, nr, nc, sigma );
	bzero( vx, nr * nc );				/* set flow to zero */
	bzero( vy, nr * nc );

	for( x = -wsize; x <= wsize; x++ )
		for( y = -wsize; y <= wsize; y++ ) {
			fprintf( stderr, "flow: %3d %3d\r", x, y ); fflush( stderr );
			if ( x == 0 && y == 0 ) continue;

			match( im1,im2, mtchim, nr, nc, x, y );			/* match for offset x,y */
			/* smooth( mtchim, mtchim, nr, nc, smsize );		/* smooth matches */
			DericheSmooth( mtchim, mtchim, nr, nc, sigma );

			setmax( mtchim, maxim, vx, vy, nr, nc, x, y );		/* winner-takes-all max */
		}
	fprintf( stderr, "\n" );

	free( maxim );
	free( mtchim );
}



/* verified optical flow */

void vflow( UCHAR *im1, UCHAR *im2, char *vx, char *vy, UCHAR *mask, int nr, int nc,
		int wsize, float sigma, float bw )
{
UCHAR *mtchim, *maxim_f, *maxim_b;
char *vx_b, *vy_b;
int x, y;

	dog( im1, im1, nr, nc, bw );
	dog( im2, im2, nr, nc, bw );

	mtchim  = (UCHAR *)malloc( nr * nc * sizeof(UCHAR) );
	maxim_f = (UCHAR *)malloc( nr * nc * sizeof(UCHAR) );
	maxim_b = (UCHAR *)malloc( nr * nc * sizeof(UCHAR) );
	vx_b    = (char  *)malloc( nr * nc * sizeof(char) );
	vy_b    = (char  *)malloc( nr * nc * sizeof(char) );

	/* initialize flow to zero offest in both directions */
	match( im1,im2, mtchim, nr, nc, 0, 0 );	/* matching */
	/* smooth( mtchim, maxim_f, nr, nc, smsize );	/* smooth matches */
	/* smooth( mtchim, maxim_b, nr, nc, smsize );	/* smooth matches */
	DericheSmooth( mtchim, maxim_f, nr, nc, sigma );
	DericheSmooth( mtchim, maxim_b, nr, nc, sigma );

	bzero( vx, nr * nc );				/* set flow to zero */
	bzero( vy, nr * nc );
	bzero( vx_b, nr * nc );
	bzero( vy_b, nr * nc );

	for( x = -wsize; x <= wsize; x++ )
		for( y = -wsize; y <= wsize; y++ ) {
			fprintf( stderr, "vflow: %3d %3d\r", x, y ); fflush( stderr );
			if ( x == 0 && y == 0 ) continue;

			match( im1,im2, mtchim, nr, nc, x, y );			/* match for offset x,y */
			/* smooth( mtchim, mtchim, nr, nc, smsize );		/* smooth matches */
			DericheSmooth( mtchim, mtchim, nr, nc, sigma );

			setmax( mtchim, maxim_f, vx, vy, nr, nc, x, y );		/* winner-takes-all max */
			setmax_b( mtchim, maxim_b, vx_b, vy_b, nr, nc, x, y );
		}

	verify( vx, vy, vx_b, vy_b, mask, nr, nc );
	fprintf( stderr, "\n" );

	free( vy_b );
	free( vx_b );
	free( maxim_b );
	free( maxim_f );
	free( mtchim );
}

void verify( char *vx, char *vy, char *vx_b, char *vy_b, UCHAR *mask, int nr, int nc )
{
int i, j, ii, jj;
int vx1, vy1, vx2, vy2;

	for( i = 0; i < nr; i++ )
		for( j = 0; j < nc; j++ ) {
			mask[i * nc + j] = 0;

			vx1 = vx[i * nc + j];
			vy1 = vy[i * nc + j];
			if ( vx1 >= 128 ) vx1 -= 256;
			if ( vy1 >= 128 ) vy1 -= 256;

			ii = i + vy1;
			jj = j + vx1;

			if ( ii < 0 || ii >= nr ) continue;
			if ( jj < 0 || jj >= nc ) continue;

			vx2 = vx_b[ii * nc + jj];
			vy2 = vy_b[ii * nc + jj];
			if ( vx2 >= 128 ) vx2 -= 256;
			if ( vy2 >= 128 ) vy2 -= 256;

			if ( ABS(vx1 + vx2) <= 2 && ABS(vy1 + vy2) <= 2 )
				mask[i * nc + j] = 255;
		}
}



void match( UCHAR *im1, UCHAR *im2, UCHAR *match, int nr, int nc, int xoffset, int yoffset )
{
int i, j;
int ii, jj;

	for( i = 0; i < nr; i++ )
		for( j = 0; j < nc; j++ ) {
			match[i * nc + j] = 0;
			ii = i + yoffset;
			jj = j + xoffset;
			if ( ii < 0 || ii >= nr ) continue;
			if ( jj < 0 || jj >= nc ) continue;

			match[i * nc + j] = 255 - ABS( im1[i * nc + j] - im2[ii * nc + jj] );			
		}
}


void smooth( UCHAR *in, UCHAR *out, int nr, int nc, int size )
{
int i, j;	/* image for loop indices */
int ii, jj;	/* averaging window indices */
int sum;	/* sum in averaging window */
float scale;	/* scale to convert sum to average */
UCHAR *temp;

	scale = 1.0 / (2 * size + 1);
	if ( (temp = (UCHAR *)malloc( nr * nc * sizeof(UCHAR) )) == NULL ) {
		fprintf( stderr, "Error:  malloc failed in smooth_im\n" );
		exit( YES );
	}


	/* first pass */
	for( i = 0; i < nr; i++ ) {
		for( j = 0; j < nc; j++ ) {
			sum = 0;
			for( jj = j - size; jj <= j + size; jj++ ) {
				if ( jj < 0 || jj >= nc ) continue;
				sum += in[i * nc + jj];
			}

			temp[i * nc + j] = (UCHAR)(sum * scale + 0.5);
		}
	}

	/* second pass */
	for( j = 0; j < nc; j++ ) {
		for( i = 0; i < nr; i++ ) {
			sum = 0;
			for( ii = i - size; ii <= i + size; ii++ ) {
				if ( ii < 0 || ii >= nr ) continue;
				sum += temp[ii * nc + j];
			}

			out[i * nc + j] = (UCHAR)(sum * scale + 0.5);
		}
	}

	free( temp );
}



void setmax( UCHAR *mtchim, UCHAR *maxim, char *vx, char*vy, int nr, int nc, int xoffset, int yoffset )
{
int i, n;

	n = nr * nc;

	for( i = 0; i < n; i++ )
		if ( mtchim[i] > maxim[i] ) {
			maxim[i] = mtchim[i];
			vx[i] = xoffset;
			vy[i] = yoffset;
		}
}


void setmax_b( UCHAR *mtchim, UCHAR *maxim, char *vx, char*vy, int nr, int nc, int xoffset, int yoffset )
{
int i, j, ii, jj;

	for( i = 0; i < nr; i++ ) {
		ii = i + yoffset;
		if ( ii < 0 || ii >= nr ) continue;
		for( j = 0; j < nc; j++ ) {
			jj = j + xoffset;
			if ( jj < 0 || jj >= nc ) continue;

			if ( mtchim[i * nc + j] > maxim[ii * nc + jj] ) {
				maxim[ii * nc + jj] = mtchim[i * nc + j];
				vx[ii * nc + jj] = -xoffset;
				vy[ii * nc + jj] = -yoffset;
			}
		}
	}
}


void scale( UCHAR *im, int nr, int nc )
{
int i, n;
int max, min;
float scale;

	max = min = im[0];
	for( i = 1; i < n; i++ ) {
		if ( im[i] < min ) min = im[i];
		if ( im[i] > max ) max = im[i];
	}

	if ( max == min ) scale = 0.0;
		else scale = 255.0 / (float)(max - min);

	for( i = 0; i < n; i++ )
		im[i] = (int)(scale * (im[i] - min) + 0.5);
}


void magnitude( char *vx, char *vy, UCHAR *mag, int nr, int nc )
{
int i, n;

	n = nr * nc;
	for( i = 0; i < n; i++ )
		mag[i] = (int)(sqrt( (float)(SQR(vx[i]) + SQR(vy[i])) ) + 0.5);

}


void gaussian( UCHAR *in, UCHAR *out, int nr, int nc, float sigma )
{
int i, j, k;	/* image for loop indices */
int ii, jj;	/* convolution window indices */
float sum;	/* sum in averaging window */
UCHAR *temp;	/* intermediate image for separable filter operation */
float *kernel;	/* gaussian kernel */
int size, nk;	/* gaussian kernel size */

	size = (int)(3.0 * sigma + 1.0);
	nk = 2 * size + 1;

	/* allocate memory of temprory image */
	if ( (temp = (UCHAR *)malloc( nr * nc * sizeof(UCHAR) )) == NULL ) {
		fprintf( stderr, "Error:  malloc failed in gaussian\n" );
		exit( YES );
	}

	/* allocate memory for gaussian kernel */
	if ( (kernel = (float *)malloc( nk * sizeof( float ) )) == NULL ) {
		fprintf( stderr, "Error:  malloc failed in gaussian\n" );
		exit( YES );
	}


	/* fill in the gaussian kernel */
	sum = 0.0;
	for( k = 0; k < nk; k++ ) {
		kernel[k] = normal( (float)(k - size), 0.0, sigma );
		sum += kernel[k];
	}
	for( k = 0; k < nk; k++ )
		kernel[k] /= sum;


	/* first pass */
	for( i = 0; i < nr; i++ ) {
		for( j = 0; j < nc; j++ ) {
			sum = 0;
			for( jj = j - size, k = 0; jj <= j + size; jj++, k++ ) {
				if ( jj < 0 || jj >= nc ) continue;
				sum += kernel[k] * in[i * nc + jj];
			}

			temp[i * nc + j] = (UCHAR)(sum + 0.5);
		}
	}

	/* second pass */
	for( j = 0; j < nc; j++ ) {
		for( i = 0; i < nr; i++ ) {
			sum = 0;
			for( ii = i - size, k = 0; ii <= i + size; ii++, k++ ) {
				if ( ii < 0 || ii >= nr ) continue;
				sum += kernel[k] * temp[ii * nc + j];
			}

			out[i * nc + j] = (UCHAR)(sum + 0.5);
		}
	}

	free( kernel );
	free( temp );
}


float normal( float x, float mu, float sigma )
{

	return (1.0 / (sqrt(2.0 * M_PI) * sigma)) * exp(-0.5 * SQR(x - mu) / SQR(sigma));
}


void dog( UCHAR *in, char *out, int nr, int nc, float sigma )
{
UCHAR *temp1, *temp2;
int i, n;
int pixel;
	n = nr * nc;

	/* allocate memory of temprory images */
	if ( (temp1 = (UCHAR *)malloc( n * sizeof(UCHAR) )) == NULL ) {
		fprintf( stderr, "Error:  malloc failed in dog\n" );
		exit( YES );
	}
	if ( (temp2 = (UCHAR *)malloc( n * sizeof(UCHAR) )) == NULL ) {
		fprintf( stderr, "Error:  malloc failed in dog\n" );
		exit( YES );
	}

	gaussian( in, temp1, nr, nc, sigma );
	gaussian( in, temp2, nr, nc, 2.0 * sigma );

	for( i = 0; i < n; i++ ) {
		pixel = temp1[i] - temp2[i] + 128;
		if ( pixel < 0 ) out[i] = 0;
		else if ( pixel > 255 ) out[i] = 255;
		else out[i] = pixel;
	}

	free( temp1 );
	free( temp2 );
}


void LogOfImage( UCHAR *in, UCHAR *out, int nr, int nc )
{
int i, n, pixel;

	n = nr * nc;
	for( i = 0; i < n; i++ ) {
		pixel = in[i];
		pixel = MAX( pixel, 1 );
		out[i] = (int)(  105.96 * flog10((float)pixel) + 0.5 );
	}
}
